#include<sstream>
#include<fstream>
#include<iostream>
#include<string>
#include<vector>
#include<math.h>
#include<time.h>
#ifdef WIN32
	#include<direct.h>
#else
	#include<sys/types.h>
	#include<sys/stat.h>
#endif

#include "option.h"
#include "util.h"

using namespace std;

double neighborhood(vector<int> left, vector<int> right) {
	int i, j = 0;
	int cnt = 0;
	for(i = 0; i < left.size(); i++) {
		while(left[i]>right[j]) {
			j++;
			if(j==right.size())
				break;
		}
		if(j==right.size())
			break;
		if(left[i]==right[j])
			cnt++;
	}
	return (double)(cnt*cnt)/(left.size()*right.size());
}

int main(int argc, char** argv) {
	vector<string> proteins;
	int num_interactions = 0;
	bool** PPIs;
	const int minD = 3;
	const int size_cut = 1;	
	vector<bool> avlvls;
	vector<bool> avhvls;
	int lcnt = 0, hcnt = 0;

	int i, j, k;
	int x, y;
	int maxl =0, suml = 0, maxh = 0, sumh = 0;
	vector< vector<int> > vls;
	time_t start, finish;

	Option op;
	try {
		op.init(argc,argv);
	} catch (Option::InvalidArgument& ex) {
		cerr << ex.what() << endl;
		exit(-1);
	} 
	
	if(op.showHelp()) {
		op.printUsage();
		return 0;
	}

	try {
		//1. get proteins first
		ifstream infile(op.getInputFile().c_str());
		ofstream outfile;
		if (!infile) {
			throw RuntimeException("Error! Input file " + op.getInputFile() + " does not exist!");
			exit(-1);
		}

		cout << endl;
		cout << "=====================================================" << endl;
		cout << "Reading Data from: " << op.getInputFile() << "; please wait..." << endl;

		string leftprotein, rightprotein, tmprightprotein;
		string interaction;
		while(!infile.eof()) {
			getline(infile, interaction);
			size_t breakpos = interaction.find(' ');
			if(breakpos==string::npos)
				breakpos = interaction.find('\t');
			if(breakpos!=string::npos) {
				leftprotein = interaction.substr(0, breakpos);
				tmprightprotein = interaction.substr(breakpos+1, interaction.length()-breakpos-1);
				if(tmprightprotein[tmprightprotein.length()-1]=='\r')
					rightprotein = tmprightprotein.substr(0, tmprightprotein.length() - 1);
				else
					rightprotein = tmprightprotein;
				for(i = 0; i < proteins.size(); i++) {
					if(proteins[i].compare(leftprotein)==0)
						break;
				}
				if(i==proteins.size())
					proteins.push_back(leftprotein);
				for(i = 0; i < proteins.size(); i++) {
					if(proteins[i].compare(rightprotein)==0)
						break;
				}
				if(i==proteins.size())
					proteins.push_back(rightprotein);
			}
		}	
		infile.clear();
		infile.close();
	
		//2. then get interactions
		infile.open(op.getInputFile().c_str());
		if (!infile) {
			throw RuntimeException("Error! Input file " + op.getInputFile() + " does not exist!");
			exit(-1);
		}
		PPIs = new bool*[proteins.size()];
		for(i = 0; i < proteins.size(); i++) {
			bool* PPI = new bool[proteins.size()];
			for(j = 0; j < proteins.size(); j++) {
				PPI[j] = 0;
			}
			PPIs[i] = PPI;
		}
		while(!infile.eof()) {
			getline(infile, interaction);
			size_t breakpos = interaction.find(' ');
			if(breakpos==string::npos)
				breakpos = interaction.find('\t');
			if(breakpos!=string::npos) {
				leftprotein = interaction.substr(0, breakpos);
				tmprightprotein = interaction.substr(breakpos+1, interaction.length()-breakpos-1);
				if(tmprightprotein[tmprightprotein.length()-1]=='\r')
					rightprotein = tmprightprotein.substr(0, tmprightprotein.length() - 1);
				else
					rightprotein = tmprightprotein;
				for(i = 0; i < proteins.size(); i++) {
					if(proteins[i].compare(leftprotein)==0)
						break;
				}
				for(j = 0; j < proteins.size(); j++) {
					if(proteins[j].compare(rightprotein)==0)
						break;
				}
				if(PPIs[i][j]==0) {
					PPIs[i][j] = 1;
					PPIs[j][i] = 1;
					num_interactions++;
				}
			}
		}
		infile.clear();
		infile.close();
		cout << "Number of proteins:\t" << proteins.size() << endl;
		cout << "Number of interactions:\t" << num_interactions << endl;
	
		//3. find maximal cliques
		cout << "=====================================================" << endl;
		cout << "Finding maximal cliques; please wait...." << endl;
		time(&start);
		int* edges = new int[proteins.size()];
		for(i = 0; i < proteins.size(); i++) {
			edges[i] = i;
		}
		int* tegs = new int[proteins.size()-1];
	
		srand(time(NULL));
		double v = rand()%10000;
		ostringstream  outfilename;
		outfilename << v;   
		outfile.open(outfilename.str().c_str());
		int numfoundcliques = BBforMC(edges, proteins.size(), PPIs, tegs, minD, proteins.size(), outfile, outfilename.str().c_str());
		delete []tegs;
		//delete []edges;
		outfile.close();
		time(&finish);
		cout << "Number of maximal cliques:\t" << numfoundcliques << endl;
		cout << "Running time for finding maximal cliques:\t" << finish-start << "  s" << endl;

		//4. split maximal cliques by neighborhood density.
		cout << "=====================================================" << endl;
		cout << "Splitting maximal cliques by neighborhood density; please wait...." << endl;

		infile.open(outfilename.str().c_str());
		if (!infile) {
			throw RuntimeException("Error! Input file " + outfilename.str() + " does not exist!");
			exit(-1);
		}
		vls.clear();
		vls.resize(numfoundcliques);
		int* avls = new int[numfoundcliques];	//available?
		for(i = 0; i < numfoundcliques; i++) {
			avls[i] = 0;
		}
		vector<int> vl;
		string proteinlist;
		int pid, listsize;
		j = 0;
		while(!infile.eof()) {
			getline(infile, proteinlist);
			stringstream ssp (stringstream::in | stringstream::out);
			ssp << proteinlist;
			ssp >> listsize;
			vl.clear();
			for(i = 0; i < abs(listsize); i++) {
				ssp >> pid;
				vl.push_back(pid);
			}
			vls[j] = vl;
			j++;
			if(j==numfoundcliques)
				break;
		}
		infile.clear();
		infile.close();
		
		remove(outfilename.str().c_str());
	
		time(&start);
		int ss;
		for(i = 0; i < vls.size(); i++) {
			for(j = i+1; j < vls.size(); j++) {
				ss = 0;
				for(x = 0; x < vls[i].size(); x++) {
					for(y = 0; y < vls[j].size(); y++) {
						if(vls[i][x]==vls[j][y]) {
							ss++;
							break;
						}
					}
				}
				double st = ((double)ss)*((double)ss) / (vls[i].size() * vls[j].size());			
				if(st >= op.getCliquesSimilarityThreshold()) {
					avls[i]++;
					avls[j]++;
				}
			}
		}
		
		vector< vector<int> > lvls;	//low neighborhood maximal cliques
		vector< vector<int> > hvls;	//high neighborhood maximal cliques
		for(i = 0; i < vls.size(); i++) {
			if(avls[i]<=op.getCliquesCutSizeThreshold()) {
				lvls.push_back(vls[i]);
				suml += vls[i].size();
				if(maxl < vls[i].size())
					maxl = vls[i].size();
			} else {
				hvls.push_back(vls[i]);
				sumh += vls[i].size();
				if(maxh < vls[i].size())
					maxh = vls[i].size();
			}
		}
		time(&finish);
		cout << "Number of low neighborhood density cliques:\t" << lvls.size() << endl;
		cout << "Average size of low neighborhood density cliques:\t" << (double)suml/lvls.size() << endl;
		cout << "Maximum size of low neighborhood density cliques:\t" << (double)maxl << endl;
		cout << "Number of high neighborhood density cliques:\t" << hvls.size() << endl;
		cout << "Average size of high neighborhood density cliques:\t" << (double)sumh/hvls.size() << endl;
		cout << "Maximum size of high neighborhood density cliques:\t" << (double)maxh << endl;
		cout << "Running time for splitting maximal cliques by neighborhood density:\t" << finish-start << "  s" << endl;


		//5. 2Core for low neighborhood density cliques
		cout << "=====================================================" << endl;
		cout << "Finding complexes in low neighborhood density cliques; please wait...." << endl;
		time(&start);
		for(i = 0; i < lvls.size(); i++) {
			vl = lvls[i];
			vector<int> tmpproteins(proteins.size());
			for(j = 0; j < proteins.size(); j++) {
				tmpproteins[j] = j;
			}
			while(true) {
				int maxdegree = -1;
				int maxindex = -1;
				for(j = 0; j < tmpproteins.size(); j++) {	// find a protein v in V - P such that adj(v, P) is max
					int degree_with_prediction = 0;
					for(k = 0; k < vl.size(); k++) {
						if(vl[k]==tmpproteins[j])	//avoid a protien in the prediction set.
							break;
						if(PPIs[tmpproteins[j]][vl[k]])
							degree_with_prediction++;
					}
					if(k==vl.size()) {
						if(degree_with_prediction>maxdegree) {
							maxdegree = degree_with_prediction;
							maxindex = j;
						}
					}
				}
				//no more proteins in remaining set
				if(maxindex==-1)
					break;
				//is still dense
				int edges_of_prediction = maxdegree;
				for(j = 0; j < vl.size(); j++) {
					for(k = j + 1; k < vl.size(); k++) {
						if(PPIs[vl[j]][vl[k]])
							edges_of_prediction++;
					}
				}
				double density = (double)edges_of_prediction * 2 / vl.size() / (vl.size() + 1);
	
				if(density<op.getDensityThreshold())
					break;
				
				//is 2 core, if yes, add into predictions else discard it.
				for(j = 0; j < vl.size(); j++) {
					if(!PPIs[vl[j]][tmpproteins[maxindex]]) {
						for(k = 0; k < vl.size(); k++) {
							if(PPIs[vl[j]][vl[k]]&&PPIs[vl[k]][tmpproteins[maxindex]]&&j!=k) {
								break;
							}
						}
						if(k==vl.size())	//not a 2core
							break;
					}
				}
				if(j==vl.size()) {	//a 2core
					vl.push_back(tmpproteins[maxindex]);
				}
				tmpproteins.erase(tmpproteins.begin() + maxindex);
			}
			lvls[i] = vl;
		}
		time(&finish);
		avlvls.resize(lvls.size());
		for(i = 0; i < lvls.size(); i++)
			avlvls[i] = true;
		for(i = 0; i < lvls.size(); i++) {
			for(j = i + 1; j < lvls.size(); j++) {
				if(avlvls[j]&&lvls[i].size()==lvls[j].size()) {
					for(x = 0; x < lvls[i].size(); x++) {
						for(y = 0; y < lvls[j].size(); y++) {
							if(lvls[i][x]==lvls[j][y])
								break;
						}
						if(y==lvls[j].size())
							break;
					}
					if(x==lvls[i].size())
						avlvls[j] = false;
				}
			}
		}
		maxl = 0;
		suml = 0;
		for(i = 0; i < lvls.size(); i++) {
			if(lvls[i].size()>size_cut&&avlvls[i]) {
				lcnt++;
				suml += lvls[i].size();
				if(maxl < lvls[i].size())
					maxl = lvls[i].size();
			}
		}
		cout << "Number of low neighborhood density complexes:\t" << lcnt << endl;
		cout << "Average size of low neighborhood density complexes:\t" << (double)suml/lcnt << endl;
		cout << "Maximum size of low neighborhood density complexes:\t" << (double)maxl << endl;
		cout << "Running time for finding complexes in low neighborhood density cliques:\t" << finish-start << "  s" << endl;
		//5. 2Core for high neighborhood density cliques
		cout << "=====================================================" << endl;
		cout << "Finding complexes in high neighborhood density cliques; please wait...." << endl;
		time(&start);
		vector< vector<int> > vlsr;

		for(i = 0; i < hvls.size(); i++) {
			sort(hvls[i].begin(), hvls[i].end());
		}
		do {
			vlsr.clear();
			vector<int> vl0;
			vector<int> vl1;
			double msimilarity;
			int mindex;
		
			for(i = 0; i < hvls.size(); i++) {
				vl0 = hvls[i];
				msimilarity = 0.0;
				mindex = i;
				for(j = 0; j < hvls.size(); j++) {
					vl1 = hvls[j];
					if(i!=j) {
						double similarity = neighborhood(vl0, vl1);
						if(similarity>msimilarity) {
							msimilarity = similarity;
							mindex = j;
						}
					}
				}

				if(msimilarity>=op.getSimilarityThreshold()) {
					vector<int> vlr;
					vlr.clear();
					vl1 = hvls[mindex];
					y = 0;
					for(x = 0; x < vl0.size(); x++) {
						while(vl0[x]>vl1[y]) {
							y++;
							if(y==vl1.size())
								break;
						}
						if(y==vl1.size())
							break;
						if(vl0[x]==vl1[y])
							vlr.push_back(vl0[x]);
					}
					vlsr.push_back(vlr);
				} else {
					vlsr.push_back(vl0);
				}
			}

			hvls.clear(); // remove same ones in vlsr
			for(i = 0; i < vlsr.size(); i++) {
				vl0 = vlsr[i];
				for(j = i + 1; j < vlsr.size(); j++) {
					vl1 = vlsr[j];
					if(vl0.size()==vl1.size()) {
						for(x=0; x < vl0.size(); x++) {
							if(vl0[x]!=vl1[x])
								break;
						}
						if(x==vl0.size())
							break;
					}
				}
				if(j==vlsr.size())
					hvls.push_back(vl0);
			}
		} while(hvls.size()<vlsr.size());

		time(&finish);
		avhvls.resize(hvls.size());
		for(i = 0; i < hvls.size(); i++)
			avhvls[i] = true;

		maxh = 0;
		sumh = 0;
		for(i = 0; i < hvls.size(); i++) {
			if(hvls[i].size()>size_cut&&avhvls[i]) {
				hcnt++;
				sumh += hvls[i].size();
				if(maxh < hvls[i].size())
					maxh = hvls[i].size();
			}
		}
		cout << "Number of high neighborhood density complexes:\t" << hcnt << endl;
		cout << "Average size of high neighborhood density complexes:\t" << (double)sumh/hcnt << endl;
		cout << "Maximum size of high neighborhood density complexes:\t" << (double)maxh << endl;
		cout << "Running time for finding complexes in high neighborhood density cliques:\t" << finish-start << "  s" << endl;
	
		vls.clear();
	
		for(i = 0; i < lvls.size(); i++) {
			if(lvls[i].size()>size_cut&&avlvls[i])
				vls.push_back(lvls[i]);
		}
		for(i = 0; i < hvls.size(); i++) {
			if(hvls[i].size()>size_cut&&avhvls[i])
				vls.push_back(hvls[i]);
		}
	
		vector<bool> avvls(vls.size());
		for(i = 0; i < vls.size(); i++)
			avvls[i] = true;

		vector<int> duplicatedPrediction;
		for(i = 0; i < vls.size(); i++) {
			for(j = i + 1; j < vls.size(); j++) {
				if(avvls[j]&&vls[i].size()==vls[j].size()) {
					for(x = 0; x < vls[i].size(); x++) {
						for(y = 0; y < vls[j].size(); y++) {
							if(vls[i][x]==vls[j][y])
								break;
						}
						if(y==vls[j].size())
							break;
					}
					if(x==vls[i].size()) {
						avvls[j] = false;
						if(j>=lcnt&&i<lcnt)
							duplicatedPrediction.push_back(i);
					}
				}
			}
		}
	
		ofstream ofile(op.getOutputFile().c_str());
		for(i = 0; i < vls.size(); i++) {
			if(avvls[i]) {
				vl = vls[i];
				ofile << ">";
				for(j = 0; j < vl.size(); j++) {
					ofile << "\t" << proteins[vl[j]];
				}
				ofile << "\n";
			}
		}
		ofile.clear();
		ofile.close();

		for(i = 0; i < proteins.size(); i++)
			delete []PPIs[i];
		delete []PPIs;

	} catch (exception &ex) {
		cerr << ex.what() << endl;
		op.printUsage();
		exit(-1);
	} 
}
