#include "option.h"

/* initialize options according to command line */
void Option::init(int argc, char** argv)
   throw (InvalidArgument)
{
	// loop through all the arguments except for the program name itself
	try {
		for (int i = 1; i < argc; i++) {
			char* p = argv[i];

			if (*p != '-') throw InvalidArgument();

			switch (p[1]) // switch on whatever comes after dash
			{
				case 'i': if (i+1 < argc) {
					input_file  = argv[++i];
					break;
				}
				case 'o': if (i+1 < argc) {
					output_file  = argv[++i];
					hasoutputfile = true;
					break;
				}
				case 'h': show_help   = true; break;
				case 't': 
					if (i+1 < argc) {
						std::istringstream iss(argv[++i]);
						iss >> tt;
						break;
					}
				case 'c':
					if (i+1 < argc) {
						std::istringstream iss(argv[++i]);
						iss >> ct;
						break;
					}
				case 'd':
					if (i+1 < argc) {
						std::istringstream iss(argv[++i]);
						iss >> dt;
						break;
					}
				case 's':
					if (i+1 < argc) {
						std::istringstream iss(argv[++i]);
						iss >> st;
						break;
					}
				default : throw InvalidArgument();
			}
		}
		if(!hasoutputfile) {
			output_file = input_file;
			output_file += ".t";
			std::ostringstream  stt;
			stt << tt;
			output_file += stt.str();
			output_file += ".c";
			std::ostringstream  sct;
			sct << ct;
			output_file += sct.str();
			output_file += ".d";
			std::ostringstream  sdt;
			sdt << dt;
			output_file += sdt.str();
			output_file += ".s";
			std::ostringstream  sst;
			sst << st;
			output_file += sst.str();
		}

		if(!hasoutputfile) {
			output_file += ".out";
		}
	} catch (...) {
		throw InvalidArgument();
	}
}


void Option::printUsage() const {
	std::cout <<
"Options:\n"
"\n"
"        -h display this help and exit\n"
"        -i input_file\n"
"        -o output_file (default: input_file.t<val>.c<val>.d<val>.s<val>.out)\n"
"        -t similarity threshold during computation of neighborhood density\n"
"           (default: 0.3)\n" 
"        -c occurrence threshold during computation of neighborhood density\n"
"           (default: 3)\n"
"        -d density threshold during computation of predicted complexes in\n"
"           regions with low neighborhood density (default: 0.7)\n"
"        -s similarity threshold during computation of predicted complexes in\n"
"           regions with high neighborhood density (default: 0.2)\n";
}
